package sam.machine.lang ; 

import java.util.* ; 
import java.io.* ; 
import sam.util.* ; 
import sam.antlr.autolog.* ;  // Compiler

   /**
     *  AutoLog Theory 
     *  Theory extends Vector<Inference> and does NOT have separate 
     *  internal constructor. Theory fields are constructed by 
     *  AutoLogAnalyzer(AA) uses theory details loaded from the editor.   
     *  @version 10/18/2022   
     *  As a test, AA calls this.showTheorySummary(), 
     *  which calls this.setArityKeyPatterns.
      */
public class Theory extends Vector<Inference> {  
   // non-generic static renaming, can use for [] and casting, w/o warnings 
   // similar device used by AutoLogAnalyer
   static class PatternVector extends Vector<Pattern> {}  
             // ^ used to create arityKeyPa tterns at line 55 below
   // Inference components are indexed in AutoLogAnalyzer
   public static int TRUE_INDEX = 0 ;  // name[0],equals"true"), etc.
   public static int GOAL_INDEX = 1 ;   
   public static int FALSE_INDEX = 2 ; 
   public static int EQ_INDEX = 3 ;  // 6/19/2021
   /// following 6 fields constructed by AutoLogAnalyzer 
   protected int count ; // how many indexed fields via AA
   public String[] name ;           // print names of functors by index
   public int[] arity ;             // arity of functors by index
   public byte[] type ;             // types of functors by index
   public Vector<Pattern>[] keyPatterns ;  // by index, set by AutoLogAnalyzer
   public Vector<Pattern>[] rewriteKeyPatterns ; // added 2/8/2021
   public Vector<Pattern>[] arityKeyPatterns;  // NEW 10/13/2022 
        
   
   /** 
     *  Compute largest arity in the theory using this.arity array,
     *  after set by AutoLogAnalyzer. 
     *      this.arityKeyPatterns.length == this.maxArity()
      */
   public int maxArity() { 
      int max = 0 ;
      for (int i=0 ; i<this.arity.length ; i++)
         if (arity[i]>max) max=arity[i] ;        
      return max ;
   } 

   /**
     *  AutoLogAnalyzer calls setArityKeyPatterns() after initial 
     *  indexing. Intended for theory that is not predicative but is indexical:
     *  !theory.isEmpty() && !theory.isPredicative() && theory.isIndexical()
      */
   public void setArityKeyPatterns() {   
      int max=this.maxArity() ; // largest arity in this Theory
      if (max==0) return ; // no argument ket patterns to set  
      // ELSE ... termporary arityKeyPatterns, static/nongeneric construction
      PatternVector[] tempAKP  =new PatternVector[max+1] ; // 0,1,2,..,max  
      for (int a=1 ; a<=max ; a++)  // <=
         tempAKP[a] = new PatternVector() ;  // initialize pattern vectors 
      // populate tempAKP by arity, tempAKP[0] stays empty
      for (int indx = 3 ; indx < count ; indx++) { // 0:true, 1:goal, 3:false have arity 0 
         int a = this.arity[indx] ;   
         if (a>0) { 
            Vector<Pattern> pats = keyPatterns[indx] ; 
            // add new pats if any -- pats is null unless AA needed to add any
            if (pats!=null ) addNew(pats,tempAKP[a]) ;    
         }  
      } // tempAKP populated 
      this.arityKeyPatterns = tempAKP ;
      // Trim all the arityKeyPatterns vectors to size...
      for (int a=0 ; a<this.arityKeyPatterns.length ; a++)
         if (this.arityKeyPatterns[a] != null) 
            this.arityKeyPatterns[a].trimToSize() ;
   } // end setArityKeyPatterns
                        
   // addNew patterns from more into already
   protected void addNew(Vector<Pattern> more, PatternVector already) { 
      Iterator it= more.iterator() ;  
      while (it.hasNext()) { 
         Pattern pat = (Pattern)it.next() ;   
            //System.out.println("pat="+pat) ;
            //System.out.println("already="+already) ;
         if (!already.contains(pat)) already.add(pat) ;  
      }
   } // end addNew

   /**
     *  Is every rule deep indexical?
     *  @version 11/21/2021 
     */
   public boolean isIndexical() {
      Iterator it = this.iterator() ; 
      while(it.hasNext()) 
         if (!((Inference)(it.next())).isIndexical()) return false ;
      return true ;
   }

   /**
     *  Is every rule deep predicative?
     *  @version 11/21/2021 
     */
   public boolean isPredicative() {
      Iterator it = this.iterator() ; 
      while(it.hasNext()) 
         if (!((Inference)(it.next())).isPredicative()) return false ;
      return true ;
   }

   /**
     *  Is every rule deep byName?
     *  @version 11/21/2021 
     */
   public boolean byName() {
      Iterator it = this.iterator() ; 
      while(it.hasNext()) 
         if (!((Inference)(it.next())).byName()) return false ;
      return true ;
   }

   /**
     *  Show Theory Summary 
     *  Called by report(Theory) in AutoLogAnalyzer
     *  first-version 11/21/2021 
     *  @version 10/17/2022  report setArityKeyPatterns when available
       */
   public StringBuffer showTheorySummary() {
      String s3="   ",
             s6="      " ;
      StringBuffer sb = new StringBuffer() ;
      sb.append("Theory Summary \n") ;
      sb.append(s3+"theory structural summary \n") ; 
         sb.append(s6+"theory.isIndexical()="+this.isIndexical()+"\n") ;
         sb.append(s6+"theory.isPredicative()="+this.isPredicative()+"\n") ;
         sb.append(s6+"theory.byName()="+this.byName()+"\n") ;
      sb.append(s3+"inference forms structural summaries (Rule/Rewrite) \n") ; 
      Iterator it = this.iterator() ; 
      int N=1 ; 
      while(it.hasNext()) {
         Inference inference = (Inference)it.next() ; 
         sb.append( s6+N+" .isIndexical()="+inference.isIndexical() ) ;
         sb.append( " .isPredicative()="+inference.isPredicative() ) ;
         sb.append( " .byName()="+inference.byName()+"\n" ) ;
         N+=1 ; 
      }          
      /* append arityKeyPatterns test/report here */
      this.setArityKeyPatterns() ; 
      sb.append("\nRule literal key patterns indexed by arity\n") ;
      sb.append(this.showAKP()) ;  
      return sb ;
   }
 
   /** 
     * Show arityKeyPatterns
     * as separate report for this.showTheorySummary() 
     * @version 10/17/2022
      */
   public StringBuffer showAKP() { 
      StringBuffer sb = new StringBuffer() ;
      sb.append("   arity:    key patterns \n") ;
      for (int a=1 ; a <= this.maxArity() ; a++) { 
         sb.append("   "+a) ; 
         sb.append(":") ;
         sb.append(" ") ;
         sb.append(AutoLogAnalyzer.string(this.arityKeyPatterns[a])) ;
         sb.append("\n") ;
      } 
      return sb ;
   } 

   /** 
     *  String display of this theory
     */
   public String toString() { 
      StringBuffer sb = new StringBuffer() ; 
      Iterator inferences = this.iterator() ;
      while (inferences.hasNext()) 
         sb.append(((Inference)inferences.next()).toString() + "\n") ;
      return sb.toString() ; 
   }

   /**
     *  Easy translate String to a theory, for testing.
      */
   public static Theory translate(String s) { 
      try { 
         InputStream is = new ByteArrayInputStream(s.getBytes()) ;
         return sam.antlr.autolog.Compiler.translate(is) ; 
      } 
      catch(Exception xxx) { 
         System.out.println(xxx) ;  // say what happened
         return null ;  // test for != null to continue 
      }
   }

   /**
     *  Testing Theory.translate ...  
     *  How can we catch BAD input gracefully?
     *  DUNNO yet.
      */
   public static void main(String[] args) { 
      String s =" " ;   // empty Vector<Inference> printed
      //String s = "X∧X=X.  p,q=>r." ; 
      //String s = "p(f(X))) (a) => goal. p=q." ;  // experimenting with an exceptions 
      Theory theory=translate(s) ; 
      System.out.println(theory) ; 
      System.out.println("theory.size="+theory.size()) ; 
      System.out.println("theory.FALSE_INDEX="+theory.FALSE_INDEX) ; 
   }
}

   //////  following 3 fields were for early dev tests
   // public HashMap<String,String> hints ;   // OLD UNUSED so far
   // public boolean subfactual ;             // OLD UNUSED so far
   // public PredicateRecord predrec ;        // OLD UNUSED so far 
   //////

